;=-------------------------------------------------------------------------=
;
;			MZ Loader 2.0 FAT16 Boot Sector
;
;     Copyright (c) 1999-2001 by Thomas Kjoernes, all rights reserved.
;
;
;	Boot Sector Overview:
;	---------------------
;
;	- Support for multiple boot images. You can add more images
;	  between the szImageListDef/End labes. The first image found
;	  in the root directory will be loaded.
;
;	- Support for "MZ" executables as well as plain binaries.
;
;
;	Known limitations:
;	------------------
;
;	- Does not support FAT12 partitions beyond 8GB (INT13 extensions).
;	- Does not provide any error messages; calls INT18 on any errors.
;
;
;	Memory Map:
;	-----------
;
;	Ŀ
;	 Interrupt Vector Table  0000
;	Ĵ
;	     BIOS Data Area      0040
;	Ĵ
;	 PrtScr Status / Unused  0050
;	Ĵ
;	   Image Load Address    0060
;	Ĵ
;	    Available Memory     nnnn
;	Ĵ
;	     2KB Boot Stack      A000 - SizeOf(BootSector) - 2KB
;	Ĵ
;	       Boot Sector       A000 - SizeOf(BootSector)
;	Ĵ
;                                  A000
;
;
;	Boot Image Startup (register values):
;	-------------------------------------
;
;	ds = load segment, this is where the executable is loaded.
;	es = free segment, very first available paragraph above image.
;	ss = this is the stack segment somewhere near top of memory.
;
;	dl = boot drive number
;
;	dh = ??? (undoc)
;	cx = ??? (undoc)
;	si = ??? (undoc)
;	di = ??? (undoc)
;
;    ax:bx = CS:IP used to RETF to the image.
;
;    ss:sp = pointer to 2KB stack.
;    ss:bp = pointer to boot sector, with MZ_LOADER_VARS located below it.
;
;	The MZ_LOADER_VARS structure includes the relative LBA address of
;	the disk data area and the image load and FAT segment addresses.
;
;=-------------------------------------------------------------------------=

	.386p

	INCLUDE	fat.inc
	INCLUDE	mzldr.inc

codeSegment SEGMENT USE16 'boot'

	ASSUME	cs:codeSegment, ds:codeSegment, ss:codeSegment

	FAT16_BOOT <"MZLDR2.0">

	;=----------------------------------------------------------=
	; Detect memory, setup stack, buffers and relocate ourself
	;=----------------------------------------------------------=

	int	12h		; get conventional memory size
	shl	ax, 6		; and convert it to paragraphs

	sub	ax, 512 / 16	; es = ax - boot sector size
	mov	es, ax		;

	; setup stack and base pointers

	sub	ax, 2048 / 16	; ss = ax - 2KB
	mov	ss, ax		;

	mov	sp, 2048	; 2KB stack
	mov	bp, sp		;

	mov	ax, 0060h	; ax = "load" segment

	; copy ourself to top of memory

	mov	cx, 256
	mov	si, 7C00h
	xor	di, di
	rep	movs WORD PTR es:[di], cs:[si]

	; jump to relocated code

	push	es
	push	OFFSET mainEntryPoint
	retf

mainEntryPoint:

	mov	BS.bsDriveNumber, dl

	mov	es, ax		; es = "load" segment

	mov	ax, BPB.bpbBytesPerSector
	shr	ax, 4

	mov	di, ss
	sub	di, ax

	;=----------------------------------------------------------=
	; Load entire Root Directory (maximum 16KB for FAT12/FAT16)
	;=----------------------------------------------------------=

	mov	ax, 32
	mov	si, BPB.bpbRootEntries
	mul	si
	div	BPB.bpbBytesPerSector

	xchg	ax, cx		; cx = root sectors, ax = zero

	mov	al, BPB.bpbNumberOfFATs
	mul	BPB.bpbSectorsPerFAT

	;* assumes root LBA is < 65536

	add	ax, BPB.bpbReservedSectors
	call	int13ReadSector

	push	dx		; mzldrDataAreaHi
	push	ax		; mzldrDataAreaLo
	push	di		; mzldrFATSegment	
	push	es		; mzldrBufSegment

	;=----------------------------------------------------------=
	; Look through list of "images" and load the first match
	;=----------------------------------------------------------=

	xchg	ax, si		; ax = root entries

	mov	si, OFFSET szImageListDef

bootFindNext:
	call	fat16FindFile	; found?
	jnc	bootLoadFile	;

	add	si, cx
	cmp	si, OFFSET szImageListEnd
	jb	bootFindNext

	int	18h

	; ToDo: add buffer overrun checking

bootLoadFile:
	call	fat16ReadCluster
	jnc	bootLoadFile

	;=----------------------------------------------------------=
	; Do some special checks on the image and transfer control
	;=----------------------------------------------------------=

	mov	ax, VAR.mzldrBufSegment
	mov	ds, ax

	xor	bx, bx
	xor	si, si

	; check for "MZ" executable signature

	cmp	WORD PTR ds:[si], "ZM"
	jne	bootExecFile

	add	ax, ds:[si][08h]	; ax = image base
	mov	cx, ds:[si][06h]	; cx = reloc items
	mov	bx, ds:[si][18h]	; bx = reloc table pointer

	jcxz	bootHackDone

bootHackFile:
	mov	di, ds:[bx][0]		; di = item ofs
	mov	dx, ds:[bx][2]		; dx = item seg (rel)
	add	dx, ax			; dx = item seg (abs)

	push	ds			;
	mov	ds, dx			; ds = dx
	add	ds:[di], ax		; fixup
	pop	ds			;

	add	bx, 4			; point to next entry
	dec	cx			; more?
	jnz	bootHackFile		;

bootHackDone:

	; adjust CS:IP according to header

	mov	bx, ds:[si][14h]	; ip
	add	ax, ds:[si][16h]	; cs

bootExecFile:

	mov	dl, BS.bsDriveNumber

	push	ax
	push	bx
	retf

;---------------------------------------------------------------------------
;	fat16ReadCluster
;
;	Entry:
;		ch = 0
;	      	si = cluster number to read
;		es = segment of read buffer
;	Exit:
;	      	si = next cluster number in chain (if CF == 0)
;		es = next segment
;
fat16ReadCluster PROC NEAR

	lea	ax, [si-2]

	; cx = sectors per cluster

	mov	cl, BPB.bpbSectorsPerCluster
	mul	cx	

	; add start of dataarea

	add	ax, VAR.mzldrDataAreaLo
	adc	dx, VAR.mzldrDataAreaHi

	call	int13ReadSector

	shr	bx, 4		; convert bytes into paragraphs

	mov	ax, es		; ax = our buffer
	add	ax, bx		; ax = old buffer + bytes read

	push	ax

	mov	ax, 2
	mul	si
	div	BPB.bpbBytesPerSector
	mov	si, dx

	mov	cl, 1
	mov	es, VAR.mzldrFATSegment
	add	ax, BPB.bpbReservedSectors
	cwd
	call	int13ReadSector

	mov	si, es:[si]

	pop	es

	cmp	si, FAT16_EOF	; EOF?
	cmc
	ret

fat16ReadCluster ENDP

;---------------------------------------------------------------------------
;	fat16FindFile
;
;	Entry:
;	      	ax = number of directory entries
;	     cs:si = pointer to filename string
;	        es = segment of root directory
;	Exit:
;		cx = 0Bh
;		si = initial cluster number
;	     es:di = pointer to matching entry (if CF == 0)
;
fat16FindFile PROC NEAR

	mov	cx, 11
	mov	dx, ax
	xor	di, di

fat16FindFileNext:
	pusha
	rep	cmps BYTE PTR cs:[si], es:[di]
	popa
	je	fat16FindFileDone

	add	di, 32
	dec	dx
	jnz	fat16FindFileNext

	stc
	ret

fat16FindFileDone:
	mov	si, es:[di].dirCluster
	ret

fat16FindFile ENDP

;---------------------------------------------------------------------------
;	int13ReadSector
;
;	Entry:
;		es = sector buffer segment
;		cx = sector count
;	     dx:ax = sector number (32-bit LBA)
;	Exit:
;		bx = number of bytes read
;	     	cx = zero
;	     dx:ax = sector number (32-bit LBA) of next sector
;
int13ReadSector PROC NEAR

	push	di

	xor	bx, bx

int13ReadNext:
	mov	di, 5		; five retries

int13ReadMore:
	pusha

	; add start of partition (harddisks only)

	add	ax, WORD PTR BPB.bpbHiddenSectors[0]
	adc	dx, WORD PTR BPB.bpbHiddenSectors[2]

	xor	cx, cx		; cx = null

	xchg	ax, cx		; ax = null, cx = LBA[0]
	xchg	ax, dx		; ax = LBA[2], dx = null

	div	BPB.bpbSectorsPerTrack

	xchg	ax, cx		; ax = LBA[0], cx = Result[2]

	div	BPB.bpbSectorsPerTrack

	xchg	dx, cx		; dx = Result[2], cx = Sector#
	inc	cx		; cx = one-based Sector#

	div	BPB.bpbHeadsPerCylinder

	mov	ch, al		; ch = Cylinder# [7:0]
	shl	ah, 6		; ah = Cylinder# [9:8]
	or	cl, ah		; combine Cylinder & Sector fields

	mov	dh, dl		; dh = Head#

	mov	dl, BS.bsDriveNumber

	mov	ax, 0201h
	int	13h
	jnc	int13ReadDone

;*	mov	ah, 00h		; uncomment if you want DISK RESETs
;*	int	13h		;

	popa

	dec	di
	jnz	int13ReadMore

	int	18h

int13ReadDone:
	popa

	add	bx, BPB.bpbBytesPerSector

	inc	ax
	jnz	int13ReadSkip
	inc	dx

int13ReadSkip:
	dec	cx
	jnz	int13ReadNext

	pop	di
	ret

int13ReadSector ENDP

; list of possible boot images

szImageListDef	LABEL	BYTE
	DB	"LOADER  EXE"
szImageListEnd	LABEL	BYTE

	BOOT_MAGIC_HERE

codeSegment ENDS

	END
